/*
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2011 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

package org.ow2.orchestra.designer.fileNavigator;

import flex.messaging.messages.AsyncMessage;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.servlet.ServletContext;
import javax.wsdl.WSDLException;
import javax.xml.bind.JAXBException;
import org.granite.context.GraniteContext;
import org.granite.gravity.Gravity;
import org.granite.messaging.webapp.HttpGraniteContext;
import org.ow2.orchestra.bpmn2bpel.BarGenerator;
import org.ow2.orchestra.designer.bpmn.model.ProcessModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.designer.models.transformers.bpmn2designer.Bpmn2DesignerTransformer;
import org.ow2.orchestra.designer.models.transformers.designer2bpmn.Designer2BpmnTransformer;
import org.ow2.orchestra.designer.server.ConcurrentProcessesManager;
import org.ow2.orchestra.designer.server.SessionsManager;
import org.ow2.orchestra.designer.wsdlNavigator.WsdlNavigator;
import org.ow2.orchestra.jaxb.bpmn.Definitions;
import org.ow2.orchestra.jaxb.bpmn.Import;
import org.ow2.orchestra.jaxb.bpmn.JaxbUtil;
import org.ow2.orchestra.util.Misc;
import org.xml.sax.SAXException;

public class Navigator {

  protected File projectsRootDirectory;

  protected static final Logger LOG = Logger.getLogger(Navigator.class.getName());
  protected String userName;
  protected String workspacesPath = null;
  protected HttpGraniteContext servletGrdsContext;
  protected Gravity gravity;

  public Navigator() {
    // search the user name; by default the user is Anonymous
    userName = new SessionsManager().getUsername();
    // create the workspace of user
    createSpaceUser();

    servletGrdsContext = (HttpGraniteContext) GraniteContext.getCurrentInstance();
    if (servletGrdsContext != null) {
      ServletContext servletContext = servletGrdsContext.getServletContext();
      gravity = (Gravity) servletContext.getAttribute("org.granite.gravity.Gravity");
    }
  }

  /**
   * create the space for the user connected
   */
  protected void createSpaceUser() {
    // add a file separator in a path
    workspacesPath = UtilMakeWorkspaces.getWorkspacesPath();

    if (!workspacesPath.endsWith(File.separator)) {
      workspacesPath += File.separator;
    }

    this.projectsRootDirectory = new File(workspacesPath + this.userName);
    // check if the workSpaces directory exist, if no make the workspaces
    if (!this.projectsRootDirectory.exists()) {
      // making of workSpaces, if failed it throws a RuntimeException
      if (!this.projectsRootDirectory.mkdirs()) {
        String message = "Unable to create workspaces directory: '" +
            this.projectsRootDirectory.getAbsolutePath() + "'. Please check directories permissions.";
        Navigator.LOG.log(Level.SEVERE, message);
        throw new RuntimeException(message);
      }
      // copy the examples in workSpaces user
      File exp = new File((this.getClass().getClassLoader().getResource("Template")).getPath());
      if (exp.exists()) {
        copy(exp, this.projectsRootDirectory);
      }
    }
  }

  /**
   * get the list of projects available in the designer.
   *
   * @return
   * @throws java.io.IOException
   */
  public List<Project> getProjectsAndProcess() throws IOException {
    // List files in the root directory (= projects)
    final File[] projectFiles = this.projectsRootDirectory.listFiles();

    //First create the owner
    final Owner owner = new Owner();
    owner.setName(userName);

    final List<Project> projects = new ArrayList<Project>();
    for (final File projectFile : projectFiles) {
      if (!projectFile.isHidden()) {
        if (projectFile.isDirectory()) {
          final Project project = new Project();
          project.setName(projectFile.getName());
          project.setOwner(owner);
          final String[] processNames = projectFile.list();
          for (final String processName : processNames) {
            final Process process = new Process();
            process.setName(processName);
            process.setProject(project);
            project.getProcesses().add(process);
          }
          projects.add(project);
        } else {
          Navigator.LOG.log(Level.WARNING, "Warning Directories on server is corrupt...");
        }
      }
    }
    owner.getProjects().addAll(projects);
    projects.addAll(this.getSharedProjectsAndProcess());

    return projects;
  }

  /**
   * @param project: project to create on disk
   * @return true if the project is successfully created
   */
  public boolean createNewProject(final Project project) {
    if (project == null || project.getName() == null || project.getName().startsWith(".")) {
      throw new IllegalArgumentException("project or project name is null or project name start with '.'");
    } else {
      final File file = new File(this.projectsRootDirectory, project.getName());
      return !file.exists() && file.mkdir();
    }
  }

  /**
   * return the access permission  of process defined by process
   *
   * @param process : a process
   * @return
   * @throws IOException
   */
  public final List<UserAccessPermissions> getAccessProperties(final Process process) throws IOException {
    List<UserAccessPermissions> lUserAccess = new ArrayList<UserAccessPermissions>();
    Properties p = new Properties();
    try {
      this.checkProcess(process);
      final File processDirectory;
      if ((process.getProject().getOwner().getName().equals(userName))) {
        final File projectDirectory = new File(this.projectsRootDirectory, process.getProject().getName());
        processDirectory = new File(projectDirectory, process.getName());
      } else {
        processDirectory = this.getDirectoryOfSharedProcess(process);
      }
      if (!new File(processDirectory, "access.properties").exists()) {
        createFileAccessProperties(processDirectory);
      }
      FileInputStream in = new FileInputStream(new File(processDirectory, "access.properties"));
      p.load(in);
      for (String userName : p.stringPropertyNames()) {
        UserAccessPermissions userAccess = new UserAccessPermissions();
        userAccess.setUser(userName);
        userAccess.updateAccess(p.getProperty(userName));
        lUserAccess.add(userAccess);
      }
      return lUserAccess;
    } catch (final IOException e) {
      this.logException(e);
      //throw e;
      return null;
    }
  }

  /**
   * Performs a load operation on a process. This method calls the
   * {@link Bpmn2DesignerTransformer#bpmnToModel(org.ow2.orchestra.jaxb.bpmn.Definitions)} on the returned
   * {@link org.ow2.orchestra.jaxb.bpmn.Definitions} retrieved by unmarshaling the the {@link Process} file.<br/>
   *
   * @param process the {@link Process} from which the model should be loaded
   * @return Returns the loaded {@link ProcessModel}
   * @throws IOException
   * @throws JAXBException
   * @throws SAXException
   * @throws TransformationException
   * @see @link Transformer#modelToBpmn(Definitions)
   * @see JaxbUtil#unmarshalBpmn(File)
   */
  public final ProcessModel openFile(final Process process) throws IOException, JAXBException, SAXException,
      TransformationException {
    try {

      File processDirectory;

      this.checkProcess(process);
      if ((process.getProject().getOwner().getName().equals(userName))) {
        final File projectDirectory = new File(this.projectsRootDirectory, process.getProject().getName());
        processDirectory = new File(projectDirectory, process.getName());
      } else {
        processDirectory = this.getDirectoryOfSharedProcess(process);
      }
      // recover the token if a writing user
      if (process.getProject().getOwner().getName().equals(userName) ||
          checkWritingPermissionAccessProcess(process)) {
        ConcurrentProcessesManager.getInstance().takeToken(process, this.userName);
      }
      final File processFile = new File(processDirectory, process.getName() + ".xml");
      final Definitions definitions = JaxbUtil.unmarshalBpmn(processFile);

      return Bpmn2DesignerTransformer.bpmnToModel(definitions);
    } catch (final JAXBException e) {
      this.logException(e);
      throw e;
    } catch (final SAXException e) {
      this.logException(e);
      throw e;
    } catch (final TransformationException e) {
      this.logException(e);
      throw e;
    } catch (final IOException e) {
      this.logException(e);
      throw e;
    } catch (final RuntimeException e) {
      this.logException(e);
      throw e;
    }

  }

  /**
   * Performs a save operation on a process. This method calls the
   * {@link Designer2BpmnTransformer#modelToBpmn(ProcessModel)} before marshaling the
   * returned {@link Definitions} object into the {@link Process} file.
   *
   * @param process      the {@link Process} into which the model should be saved
   * @param processModel the {@link ProcessModel} to save
   * @throws IOException
   * @throws JAXBException
   * @throws SAXException
   * @throws TransformationException
   * @see Designer2BpmnTransformer#modelToBpmn(ProcessModel)
   * @see JaxbUtil#marshalBpmn(Definitions, File)
   */
  public final void saveProcess(final Process process, final ProcessModel processModel) throws IOException, JAXBException, SAXException,
      TransformationException {
    final File processDirectory;
    try {
      this.checkProcess(process);
      if ((process.getProject().getOwner().getName().equals(userName))) {
        final File projectDirectory = new File(this.projectsRootDirectory, process.getProject().getName());
        processDirectory = new File(projectDirectory, process.getName());

        if (!processDirectory.exists()) {
          processDirectory.mkdirs();
          createFileAccessProperties(processDirectory);
        }
      } else {
        // check if the current user have the writing access permission
        if (!checkWritingPermissionAccessProcess(process)) {
          throw new RuntimeException("Permission denied");
        }

        processDirectory = this.getDirectoryOfSharedProcess(process);
      }

      if (process.getProject().getOwner().getName().equals(userName) || checkWritingPermissionAccessProcess(process)) {
        String userAccess = ConcurrentProcessesManager.getInstance().checkWhoTakeToken(process);
        if ((userAccess != null && !userAccess.equals(this.userName)))/* ||
                 ConcurrentProcessesManager.getInstance().checkReadingUser(process, this.userName)) */ {
          throw new RuntimeException("Process already open by another user. Your modifications haven't been saved.");
        }
      }

      final File processFile = new File(processDirectory, process.getName() + ".xml");

      final Definitions definitions = Designer2BpmnTransformer.modelToBpmn(processModel);
      JaxbUtil.marshalBpmn(definitions, processFile);

      // copy wsdls to process directory
      WsdlNavigator wsdlNavigator = new WsdlNavigator();
      for (Import bpmnImport : definitions.getImports()) {
        if (bpmnImport.getImportType().equals("http://schemas.xmlsoap.org/wsdl/")) {
          File importedFile = new File(processDirectory, bpmnImport.getLocation());
          if (!importedFile.exists()) {
            wsdlNavigator.copyWSDL(bpmnImport.getLocation(), processDirectory);
          }
        }
      }

    } catch (final JAXBException e) {
      this.logException(e);
      throw e;
    } catch (final SAXException e) {
      this.logException(e);
      throw e;
    } catch (final TransformationException e) {
      this.logException(e);
      throw e;
    } catch (final IOException e) {
      this.logException(e);
      throw e;
    } catch (final RuntimeException e) {
      this.logException(e);
      throw e;
    }
  }

  /**
   * Check that a process is valid (a valid process has a name, a project and its project has a name)
   * before trying to open, save or export it.
   *
   * @param process The process to check
   */
  private void checkProcess(final Process process) {
    if (process == null) {
      throw (new IllegalArgumentException("process is null"));
    } else if (process.getName() == null) {
      throw (new IllegalArgumentException("process name is null"));
    } else if (process.getProject() == null) {
      throw (new IllegalArgumentException("project is null"));
    } else if (process.getProject().getName() == null) {
      throw (new IllegalArgumentException("project name is null"));
    } else if (process.getProject().getOwner() == null) {
      throw (new IllegalArgumentException("owner is null"));
    } else if (process.getProject().getOwner().getName() == null) {
      throw (new IllegalArgumentException("owner's name is null"));
    }
  }

  private void logException(final Exception e) {
    if (Navigator.LOG.isLoggable(Level.FINE)) {
      Misc.fastDynamicLog(Navigator.LOG, Level.FINE, "%s\n%s", e.getMessage(), Misc.getStackTraceFrom(e.getStackTrace()));
    } else {
      Misc.fastDynamicLog(Navigator.LOG, Level.SEVERE, "%s", e.getMessage());
    }
  }

  /**
   * Performs a export operation on a process.
   *
   * @param process the {@link Process} into which the model should be exported to BPEL
   * @throws WSDLException
   * @throws SAXException
   * @throws IOException
   * @throws JAXBException
   */
  public final void exportProcess(final Process process) throws JAXBException, IOException, SAXException, WSDLException {
    try {
      this.checkProcess(process);

      final File projectDirectory;
      final File processDirectory;
      this.checkProcess(process);
      if ((process.getProject().getOwner().getName().equals(userName))) {
        projectDirectory = new File(this.projectsRootDirectory, process.getProject().getName());
        processDirectory = new File(projectDirectory, process.getName());
      } else {
        processDirectory = this.getDirectoryOfSharedProcess(process);
        if (checkWritingPermissionAccessProcess(process))
          throw new RuntimeException("Writing permission required to export a process");
      }

      final File processFile = new File(processDirectory, process.getName() + ".xml");
      final File barFile = new File(processDirectory, process.getName() + ".bar");

      final byte[] out = new BarGenerator().generateBarStream(processFile.toURI().toURL());
      Misc.write(out, barFile);
    } catch (final MalformedURLException e) {
      this.logException(e);
      throw e;
    } catch (final JAXBException e) {
      this.logException(e);
      throw e;
    } catch (final IOException e) {
      this.logException(e);
      throw e;
    } catch (final SAXException e) {
      this.logException(e);
      throw e;
    } catch (final WSDLException e) {
      this.logException(e);
      throw e;
    } catch (final RuntimeException e) {
      this.logException(e);
      throw e;
    }
  }

  /**
   * @param process
   * @return
   * @throws IOException
   */
  public final byte[] getExportedProcess(final Process process) throws IOException {
    try {
      final File projectDirectory;
      final File processDirectory;
      this.checkProcess(process);
      if ((process.getProject().getOwner().getName().equals(userName))) {
        projectDirectory = new File(this.projectsRootDirectory, process.getProject().getName());
        processDirectory = new File(projectDirectory, process.getName());
      } else {
        processDirectory = this.getDirectoryOfSharedProcess(process);
      }

      final File barFile = new File(processDirectory, process.getName() + ".bar");
      return Misc.getAllContentFrom(barFile);
    } catch (final IOException e) {
      this.logException(e);
      throw e;
    }
  }

  /**
   * delete the directory associate with the process
   *
   * @param item : the item (process or project) to delete
   * @return true if the folder corresponding is deleted
   * @throws java.io.IOException
   */
  public boolean deleteItem(final Item item) throws IOException {
    if (item instanceof Process) {
      final Process process = (Process) item;
      if ((process.getProject().getOwner().getName().equals(userName))) {
        final File projectDirectory = new File(this.projectsRootDirectory, process.getProject().getName());

        final File processDirectory = new File(projectDirectory, process.getName());
        return Misc.deleteDir(processDirectory);
      } else {
        return checkDeletingPermissionAccessProcess(process) &&
            Misc.deleteDir(this.getDirectoryOfSharedProcess(process));
      }

    } else if ((item instanceof Project) && !item.getName().equals("Shared Processes")) {
      final Project project = (Project) item;
      final File projectDirectory = new File(this.projectsRootDirectory, project.getName());
      return Misc.deleteDir(projectDirectory);
    } else {
      return false;
    }
  }

  /**
   * upgrade the list of accesses permissions of process in parameter with the list in parameter
   *
   * @param process
   * @param accessPermission
   * @return
   */
  public boolean modifiedAccessProperties(Process process, List<UserAccessPermissions> accessPermission) {
    try {
      final File processFile;
      final File processDirectory;
      this.checkProcess(process);
      if ((process.getProject().getOwner().getName().equals(userName))) {
        final File projectDirectory = new File(this.projectsRootDirectory, process.getProject().getName());
        processDirectory = new File(projectDirectory, process.getName());
      } else {
        processDirectory = getDirectoryOfSharedProcess(process);
      }

      processFile = new File(processDirectory, "access.properties");
      Properties p = new Properties();
      FileInputStream in = new FileInputStream(processFile);
      OutputStream out = new FileOutputStream(processFile);
      p.load(in);
      for (UserAccessPermissions userAccess : accessPermission) {
        String[] tab = userAccess.toString().split(":");
        p.setProperty(userAccess.getUser(), tab[1]);
        if (gravity != null) {
          AsyncMessage message = new AsyncMessage();
          message.setDestination("gravity");
          message.setHeader(AsyncMessage.SUBTOPIC_HEADER, userAccess.getUser());
          if (userAccess.getAdmin()) {
            message.setBody("You have just had admin access on process " + process.getName() + " by " + userName);
          } else if (userAccess.getDeleting() && !userAccess.getAdmin()) {
            message.setBody("You have just had delete access on process " + process.getName() + " by " + userName);
          } else if (userAccess.getWriting() && !userAccess.getDeleting()) {
            message.setBody("You have just had write access on process " + process.getName() + " by " + userName);
          } else if (userAccess.getReading() && !userAccess.getWriting()) {
            message.setBody("You have just had read only access on process " + process.getName() + " by " + userName);
          }
          gravity.publishMessage(message);
        }
      }
      p.store(out, getComments());
    } catch (IOException e) {
      logException(e);
    }
    return true;
  }

  /**
   * copy the content of file or directory on the destination file or directory
   *
   * @param src
   * @param dest
   */
  private void copy(final File src, File dest) {

    try {
      for (Map.Entry<String, byte[]> e : Misc.getDirContentFrom(src).entrySet()) {
        File file = new File(dest, e.getKey());
        Misc.write(e.getValue(), file);
      }
    } catch (IOException e) {
      this.logException(e);
    }
  }

  /**
   * create a access permission file for process, by default the current user has all permissions
   *
   * @param sRep : a directory which content the process
   * @return : true if access.properties is created, false else
   * @throws IOException
   */
  private boolean createFileAccessProperties(File sRep) throws IOException {
    FileOutputStream out;
    File fAccess = new File(sRep, "access.properties");
    if (!fAccess.exists()) {
      // create the access file
      fAccess.createNewFile();
    }
    out = new FileOutputStream(fAccess);

    Properties p = new Properties();
    p.store(out, getComments());
    return true;
  }

  /**
   * Check if the current user has the reading permission on given process
   *
   * @param process the given process
   * @return : true if reading permission, false else.
   * @throws IOException
   */
  private boolean checkReadingPermissionAccessProcess(final Process process) throws IOException {
    UserAccessPermissions userAccessPrc = this.getUserAccessPermissions(process);
    return userAccessPrc != null && userAccessPrc.getReading();
  }

  /**
   * return the value of admin access permission, if no entries of list or the user not exist in list
   * it return  "false".
   *
   * @param process the given process
   * @return
   * @throws IOException
   */
  private boolean checkAdminPermissionAccessProcess(final Process process) throws IOException {
    UserAccessPermissions userAccessPrc = getUserAccessPermissions(process);
    return userAccessPrc != null && userAccessPrc.getAdmin();
  }


  /**
   * Check if the current user has the reading deleting on given process
   *
   * @param process the given process
   * @return : true if reading permission, false else.
   * @throws IOException
   */
  private boolean checkDeletingPermissionAccessProcess(final Process process) throws IOException {
    UserAccessPermissions userAccessPrc = getUserAccessPermissions(process);
    return userAccessPrc != null && userAccessPrc.getDeleting();
  }

  /**
   * Check if the current user has the reading permission on given process
   *
   * @param process the given process
   * @return : true if reading permission, false else.
   * @throws IOException
   */
  private boolean checkWritingPermissionAccessProcess(final Process process) throws IOException {
    UserAccessPermissions userAccessPrc = getUserAccessPermissions(process);
    return userAccessPrc != null && userAccessPrc.getWriting();
  }

  /**
   * Check if the current user has reading permission on given process
   *
   * @param process the given process
   * @return true if reading permission.
   * @throws IOException
   */
  private UserAccessPermissions getUserAccessPermissions(final Process process) throws IOException {
    // all permissions are initialized at false
    UserAccessPermissions userAccessPrc = new UserAccessPermissions();
    userAccessPrc.setUser(this.userName);
    File accessProps = new File(this.getDirectoryOfSharedProcess(process), "access.properties");
    if (accessProps.exists()) {
      Properties p = new Properties();
      FileInputStream in = new FileInputStream(accessProps);
      p.load(in);
      if (!p.isEmpty()) {
        if (p.get(this.userName) != null) {
          userAccessPrc.setUser(this.userName);
          userAccessPrc.updateAccess((String) p.get(this.userName));

        }
      }
    }
    return userAccessPrc;
  }

  /**
   * @return the list of projects available in the designer and the current has the
   *         possibility to access.
   */
  private List<Project> getSharedProjectsAndProcess() {
    final List<Project> sharedProjects = new ArrayList<Project>();
    // List files in the root directory (= projects)
    final File[] workSpacesUserFiles = this.projectsRootDirectory.getParentFile().listFiles();

    for (final File workspaceUserFile : workSpacesUserFiles) {
      if (!workspaceUserFile.isHidden()) {
        if (workspaceUserFile.isDirectory()) {
          if (!workspaceUserFile.getName().equals(this.userName)) {
            final Owner owner = new Owner();
            owner.setName(workspaceUserFile.getName());
            final File[] pjtFiles = workspaceUserFile.listFiles();
            for (final File pjtFile : pjtFiles) {
              final Project project = new Project();
              project.setOwner(owner);
              project.setName(pjtFile.getName());
              final String[] processNames = pjtFile.list();
              for (final String processName : processNames) {
                try {
                  final Process process = new Process();
                  process.setName(processName);
                  process.setProject(project);
                  if (checkReadingPermissionAccessProcess(process)) {
                    project.getProcesses().add(process);
                  }
                } catch (IOException e) {
                  this.logException(e);
                }
              }
              if (project.getProcesses().size() > 0) {
                owner.getProjects().add(project);
                sharedProjects.add(project);
              }
            }
          }
        } else {
          Navigator.LOG.log(Level.WARNING, "Warning Directories on server is corrupt...");
        }
      }
    }
    return sharedProjects;
  }

  /**
   * return the directory of process in parameter
   *
   * @param process
   * @return
   */
  private File getDirectoryOfSharedProcess(Process process) {
    return new File(this.workspacesPath + File.separator + process.getProject().getOwner().getName() +
        File.separator + process.getProject().getName() + File.separator + process.getName());
  }

  /**
   * return the list of user in designer orchestra
   *
   * @return the list of user of designer
   */
  public List<String> getListUserOfDesigner() {
    final List<String> result = new ArrayList<String>();
    for (File file : this.projectsRootDirectory.getParentFile().listFiles()) {
      if (file.isDirectory()) {
        result.add(file.getName());
      }
    }
    return result;
  }

  /**
   * get the comment for the files permissions accesses
   *
   * @return
   */
  private String getComments() {
    return "#access properties of process \n#format: name_user=read;write;delete;admin \n#access= true if permit and  " +
        "false else \n";
  }


  /**
   * return the list of processes which can to manage the access permission
   * for the current user
   *
   * @return
   * @throws IOException
   */
  public List<Process> getOwnAdminProcess() throws IOException {
    List<Process> result = new LinkedList<Process>();
    for (Project proj : this.getProjectsAndProcess()) {
      for (Process proc : proj.getProcesses()) {
        if (proc.getProject().getOwner().getName().equals(userName)) {
          result.add(proc);
        } else {
          if (this.checkAdminPermissionAccessProcess(proc)) {
            result.add(proc);
          }
        }
      }
    }
    return result;
  }

  /**
   * determine if the token of process is free or no
   *
   * @param process
   * @return : return the login of user who have the token.
   */
  public String whoTakeToken(Process process) {
    return ConcurrentProcessesManager.getInstance().checkWhoTakeToken(process);
  }

  /**
   * Free a token for a given process.
   *
   * @param process The given process
   */
  public void freeTokenForProcess(Process process) {
    Navigator.LOG.log(Level.FINE, "free (closing process) by  : " + this.userName);
    ConcurrentProcessesManager.getInstance().freeToken(process, this.userName);
  }

  /**
   * refresh tokens on processes obtained by user.
   */
  public void clientPing() {
    Navigator.LOG.log(Level.FINE, "ping receive from: " + this.userName);
  }
}

