package org.ow2.orchestra.designer.wsdlNavigator;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.wsdl.Definition;
import javax.wsdl.Import;
import javax.wsdl.Message;
import javax.wsdl.Operation;
import javax.wsdl.Part;
import javax.wsdl.PortType;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.operations.InterfaceModel;
import org.ow2.orchestra.designer.bpmn.model.operations.OperationModel;
import org.ow2.orchestra.designer.fileNavigator.Navigator;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.wsdl.WsdlInlineUtil;
import org.ow2.orchestra.util.wsdl.WsdlUtil;

public class WsdlNavigator {

  private final File projectsRootDirectory;

  private static final Logger LOG = Logger.getLogger(WsdlNavigator.class.getName());

  public WsdlNavigator() {
    // Get web app classes directory (/WEB-INF/classes)
    final File applicationDir;
    try {
      applicationDir = new File(WsdlNavigator.getAppPath().toURI());
    } catch (final URISyntaxException e) {
      throw new RuntimeException(e);
    }

    // get web app root dir
    final File warRootDir = applicationDir.getParentFile().getParentFile();

    // create designer files directory
    this.projectsRootDirectory = new File(warRootDir, "OrchestraDesignerWSDLs");
    if (!this.projectsRootDirectory.exists()) {
      this.projectsRootDirectory.mkdir();
    }
  }

  /**
   * @return the list of projects available in the designer.
   */
  public List<Service> getAvailableServices() {
    // List files in the root directory (= projects)
    final File[] wsdlFiles = this.projectsRootDirectory.listFiles();

    final List<Service> services = new ArrayList<Service>();

    for (final File wsdlFile : wsdlFiles) {
      if (!wsdlFile.isHidden()) {
        try {
          Definition wsdlDefinition = WsdlUtil.readWsdl(wsdlFile.toURI().toURL());
          for (Object portType : wsdlDefinition.getPortTypes().values()) {
            Service service = new Service();
            service.setName(((PortType) portType).getQName().getLocalPart());
            service.setNamespace(((PortType) portType).getQName().getNamespaceURI());
            service.setWsdlPath(wsdlFile.getName());
            InterfaceModel interfaceModel = new InterfaceModel();
            interfaceModel.setId("_" + UUID.randomUUID().toString());
            interfaceModel.setName(((PortType) portType).getQName().getLocalPart());
            interfaceModel.setPortTypeQName(((PortType) portType).getQName().toString());
            service.setInterfaceModel(interfaceModel);
            for (Object operation : ((PortType) portType).getOperations()) {
              Operation op = (Operation) operation;
              OperationModel operationModel = new OperationModel();
              operationModel.setId("_" + UUID.randomUUID().toString());
              operationModel.setName(op.getName());
              Message wsdlInMessage = op.getInput().getMessage();
              final QName inMessageType = getMessageType(wsdlInMessage);
              if (inMessageType != null) {
                operationModel.setInMessageType(inMessageType.toString());
              }
              if (op.getOutput() != null) {
                Message wsdlOutMessage = op.getOutput().getMessage();
                final QName outMessageType = getMessageType(wsdlOutMessage);
                if (outMessageType != null) {
                  operationModel.setOutMessageType(outMessageType.toString());
                }

              }
              operationModel.setInterfaceModel(interfaceModel);
              interfaceModel.getOperations().put(op.getName(), operationModel);
            }
            services.add(service);
          }
        } catch (Exception e) {
          Misc.fastDynamicLog(LOG, Level.WARNING, "exception while reading %s: %s", wsdlFile.getPath(), e.getMessage());
        }
      }
    }
    return services;
  }

  private QName getMessageType(Message wsdlMessage) {
    final QName messageType;
    if (wsdlMessage.getParts().size() != 1) {
      messageType = null;
      Misc.fastDynamicLog(LOG, Level.WARNING,
              "Unsupported message %s: the message has %s part(s), only messages with 1 part are supported.",
              wsdlMessage.getQName(), wsdlMessage.getParts().size());
    } else {
      Part inMessagePart = (Part) wsdlMessage.getParts().values().iterator().next();
      if (inMessagePart.getElementName() != null) {
        messageType = inMessagePart.getElementName();
      } else {
        messageType = inMessagePart.getTypeName();
      }
    }
    return messageType;
  }

  public boolean uploadWsdl(Map<String, byte[]> files) throws IOException {
    Map<String, Definition> definitions = new HashMap<String, Definition>();
    Map<String, String> definitionAliases = new HashMap<String, String>();
    for (Entry<String, byte[]> file : files.entrySet()) {
      if (file.getKey().endsWith(".wsdl")) {
        // Try to parse wsdl
        Definition wsdlDefinition = WsdlUtil.readWsdl(new Locator(file.getKey(), files));
        // Inline XSD schema definitions
        WsdlInlineUtil.inlineWsdlFile(wsdlDefinition);
        // rename imports and add to definitions
        this.renameWsdlImports(file.getKey(), wsdlDefinition, definitionAliases, definitions);
      }
    }
    for (Entry<String, Definition> file : definitions.entrySet()) {
      Misc.write(WsdlUtil.writeWsdl(file.getValue()), new File(this.projectsRootDirectory, definitionAliases.get(file.getKey())));
    }
    return true;
  }

  public boolean copyWSDL(String wsdlFile, File outputDirectory) throws IOException {
    Map<String, Definition> definitions = new HashMap<String, Definition>();
    Map<String, String> definitionAliases = new HashMap<String, String>();
    // don't rename main wsdl
    definitionAliases.put(wsdlFile, wsdlFile);
    // Try to parse wsdl
    Definition wsdlDefinition = WsdlUtil.readWsdl(new File(projectsRootDirectory, wsdlFile).toURI().toURL());
    // Inline XSD schema definitions
    WsdlInlineUtil.inlineWsdlFile(wsdlDefinition);
    // rename imports and add to definitions
    this.renameWsdlImports(wsdlFile, wsdlDefinition, definitionAliases, definitions);
    for (Entry<String, Definition> file : definitions.entrySet()) {
      Misc.write(WsdlUtil.writeWsdl(file.getValue()), new File(outputDirectory, definitionAliases.get(file.getKey())));
    }
    return true;
  }

  private void renameWsdlImports(String name, Definition wsdlDefinition, Map<String, String> aliases, Map<String, Definition> definitions) {
    if (definitions.containsKey(name)) {
      return;
    }
    // Add to definitions
    definitions.put(name, wsdlDefinition);
    // Generate alias
    if (!aliases.containsKey(name)) {
      aliases.put(name, UUID.randomUUID().toString() + ".wsdl");
    }
    // Process wsdl imports
    final Map<String, List<Import>> imports = wsdlDefinition.getImports();
    if (!imports.isEmpty()) {
      for (final List<Import> l : imports.values()) {
        for (final Import impor : l) {
          // If imported definition is imported from a file,
          // rename the imports of the imported definition
          // and relocate this import with the new alias for the imported definition.
          final Definition importedDefinition = impor.getDefinition();
          String importName = impor.getLocationURI();
          if (!(importName.startsWith("http://") || importName.startsWith("https://"))) {
            if (importName.contains("/")) {
              importName = importName.substring(importName.lastIndexOf("/") + 1);
            }
            this.renameWsdlImports(importName, importedDefinition, aliases, definitions);
            impor.setLocationURI(aliases.get(importName));
          }
        }
      }
    }
  }

  public static URL getAppPath() {
    // Get Navigator classloader
    final ClassLoader loader = Navigator.class.getClassLoader();

    final String path = Navigator.class.getName().replace(".", "/") + ".class";

    // load class resource
    final java.net.URL url = loader.getResource(path);

    //b get resource path on the file system
    final String realPath = url.toExternalForm();

    // get application base directory (by removing the class path)
    final int pos = realPath.indexOf(path);
    String basePath = realPath.substring(0, pos - 1);

    // if file is in a jar, get the jar directory
    if (basePath.endsWith("!")) {
      basePath = basePath.substring(0, realPath.lastIndexOf('/'));
    }

    // return url
    try {
      return new URL(basePath);
    } catch (final MalformedURLException e) {
      throw new RuntimeException(e);
    }
  }

}
